@@ -74,9 +74,9 @@ class WithNorm(nn.Module):
         super().__init__()
         self.norm = norm_class(dim)
         self.fn = fn
-    def forward(self, x):
+    def forward(self, x, **kwargs):
         x = self.norm(x)
-        return self.fn(x)
+        return self.fn(x, **kwargs)
 
 class Chunk(nn.Module):
     def __init__(self, chunks, fn, along_dim = -1):
@@ -89,20 +89,6 @@ class Chunk(nn.Module):
         chunks = x.chunk(self.chunks, dim = self.dim)
         return torch.cat([self.fn(c) for c in chunks], dim = self.dim)
 
-class SettableArgs(nn.Module):
-    def __init__(self, fn, *args, **kwargs):
-        super().__init__()
-        self.args = args
-        self.kwargs = kwargs
-        self.fn = fn
-
-    def set_args(self, *args, **kwargs):
-        self.args = args
-        self.kwargs = kwargs
-
-    def forward(self, x):
-        return self.fn(x, *self.args, **self.kwargs)
-
 # LSH attention as described in https://openreview.net/pdf?id=rkgNKkHtvB
 # adapted from trax, stripped to what paper said needed to work
 # namely that buckets need to be at least 64 with 8 rounds of hashing
@@ -113,6 +99,7 @@ class LSHAttention(nn.Module):
                   dropout = 0.,
                   bucket_size = 64,
                   n_hashes = 8,
+                  n_local_attn_hashes = 0,
                   causal = False,
                   allow_duplicate_attention = True,
                   attend_across_buckets = True,
@@ -132,9 +119,11 @@ class LSHAttention(nn.Module):
             ' is not implemented.')
 
         self.causal = causal
-        self.n_hashes = n_hashes
         self.bucket_size = bucket_size
 
+        self.n_hashes = n_hashes
+        self.n_local_attn_hashes = n_local_attn_hashes
+
         self._allow_duplicate_attention = allow_duplicate_attention
         self._attend_across_buckets = attend_across_buckets
         self._rehash_each_round = rehash_each_round
@@ -200,10 +189,17 @@ class LSHAttention(nn.Module):
         n_buckets = seqlen // self.bucket_size
 
         buckets = self.hash_vectors(n_buckets, qk)
+
         # We use the same vector as both a query and a key.
         assert int(buckets.shape[1]) == self.n_hashes * seqlen
 
-        ticker = torch.arange(self.n_hashes * seqlen, device=device).unsqueeze(0).expand_as(buckets)
+        if self.n_local_attn_hashes > 0:
+            local_buckets = torch.full((batch_size, seqlen * self.n_local_attn_hashes), n_buckets, device=device, dtype=torch.long)
+            buckets = torch.cat((buckets, local_buckets), dim=1)
+
+        total_hashes = self.n_hashes + self.n_local_attn_hashes
+
+        ticker = torch.arange(total_hashes * seqlen, device=device).unsqueeze(0).expand_as(buckets)
         buckets_and_t = seqlen * buckets + (ticker % seqlen)
         buckets_and_t = buckets_and_t.detach()
 
@@ -221,7 +217,7 @@ class LSHAttention(nn.Module):
         sv = batched_index_select(v, st)
 
         # Split off a "bin" axis so that attention only occurs within chunks.
-        chunk_size = self.n_hashes * n_buckets
+        chunk_size = total_hashes * n_buckets
         bq_t = bkv_t = torch.reshape(st, (batch_size, chunk_size, -1))
         bqk = torch.reshape(sqk, (batch_size, chunk_size, -1, dim))
         bv = torch.reshape(sv, (batch_size, chunk_size, -1, dim))
@@ -298,22 +294,22 @@ class LSHAttention(nn.Module):
                 locs1 = buckets * chunk_size + locs1
                 locs2 = buckets * chunk_size + locs2
             locs = torch.cat([
-                torch.reshape(locs1, (batch_size, self.n_hashes, seqlen)),
-                torch.reshape(locs2, (batch_size, self.n_hashes, seqlen)),
+                torch.reshape(locs1, (batch_size, total_hashes, seqlen)),
+                torch.reshape(locs2, (batch_size, total_hashes, seqlen)),
             ], 1).permute((0, 2, 1))
 
             slocs = batched_index_select(locs, st)
-            b_locs = torch.reshape(slocs, (batch_size, chunk_size, -1, 2 * self.n_hashes))
+            b_locs = torch.reshape(slocs, (batch_size, chunk_size, -1, 2 * total_hashes))
 
-            b_locs1 = b_locs[:, :, :, None, :self.n_hashes]
+            b_locs1 = b_locs[:, :, :, None, :total_hashes]
 
-            bq_locs = b_locs1.expand(b_locs.shape[:3] + (2, self.n_hashes))
+            bq_locs = b_locs1.expand(b_locs.shape[:3] + (2, stotal_hashes))
             bq_locs = torch.reshape(bq_locs, b_locs.shape)
             bkv_locs = look_one_back(b_locs)
 
             dup_counts = (bq_locs[:, :, :, None, :] == bkv_locs[:, :, None, :, :])
             # for memory considerations, chunk summation of last dimension for counting duplicates
-            dup_counts = chunked_sum(dup_counts, chunks=(self.n_hashes * batch_size))
+            dup_counts = chunked_sum(dup_counts, chunks=(total_hashes * batch_size))
             dup_counts = dup_counts.detach()
             assert dup_counts.shape == dots.shape
             dots = dots - torch.log(dup_counts + 1e-9)
@@ -344,8 +340,8 @@ class LSHAttention(nn.Module):
                 return so_grad, slogits_grad
 
         o, logits = UnsortLogits.apply(so, slogits)
-        o = torch.reshape(o, (batch_size, self.n_hashes, seqlen, dim))
-        logits = torch.reshape(logits, (batch_size, self.n_hashes, seqlen, 1))
+        o = torch.reshape(o, (batch_size, total_hashes, seqlen, dim))
+        logits = torch.reshape(logits, (batch_size, total_hashes, seqlen, 1))
 
         if query_len != seqlen:
             query_slice = (slice(None), slice(None), slice(0, query_len))
@@ -359,11 +355,11 @@ class LSHAttention(nn.Module):
         # return unsorted attention weights
         if self._return_attn:
             attn_unsort = ((bq_t * seqlen)[:, :, :, None] + bkv_t[:, :, None, :])
-            attn_unsort = attn_unsort.view(batch_size * self.n_hashes, -1).long()
-            unsorted_dots = torch.zeros(batch_size * self.n_hashes, seqlen * seqlen, device=device)
+            attn_unsort = attn_unsort.view(batch_size * total_hashes, -1).long()
+            unsorted_dots = torch.zeros(batch_size * total_hashes, seqlen * seqlen, device=device)
             unsorted_dots.scatter_add_(1, attn_unsort, dots.view_as(attn_unsort))
             del attn_unsort
-            unsorted_dots = unsorted_dots.reshape(batch_size, self.n_hashes, seqlen, seqlen)
+            unsorted_dots = unsorted_dots.reshape(batch_size, total_hashes, seqlen, seqlen)
             attn = torch.sum(unsorted_dots[:, :, 0:query_len, :] * probs, dim=1)
 
         # return output, attention matrix, and bucket distribution
@@ -372,9 +368,10 @@ class LSHAttention(nn.Module):
 # simple full attention
 
 class FullQKAttention(nn.Module):
-    def __init__(self, causal = False):
+    def __init__(self, causal = False, dropout = 0.):
         super().__init__()
         self.causal = causal
+        self.dropout = nn.Dropout(dropout)
 
     def forward(self, qk, v, query_len = None, input_mask = None, input_attn_mask = None):
         b, seq_len, dim = qk.shape
@@ -407,6 +404,8 @@ class FullQKAttention(nn.Module):
             dot[:, i, j] = masked_value
 
         dot = dot.softmax(dim=-1)
+        dot = self.dropout(dot)
+
         out = torch.einsum('bij,bje->bie', dot, v)
 
         return out, dot, torch.empty(0)
@@ -414,7 +413,7 @@ class FullQKAttention(nn.Module):
 # Shared qk attention, using either full or LSH attention
 
 class LSHSelfAttention(nn.Module):
-    def __init__(self, dim, heads = 8, bucket_size = 64, n_hashes = 8, causal = False, attn_chunks = 1, random_rotations_per_head = False, attend_across_buckets = True, allow_duplicate_attention = True, num_mem_kv = 0, one_value_head = False, use_full_attn = False, full_attn_thres = None, return_attn = False, **kwargs):
+    def __init__(self, dim, heads = 8, bucket_size = 64, n_hashes = 8, n_local_attn_hashes = 0, causal = False, attn_chunks = 1, random_rotations_per_head = False, attend_across_buckets = True, allow_duplicate_attention = True, num_mem_kv = 0, one_value_head = False, use_full_attn = False, full_attn_thres = None, return_attn = False, post_attn_dropout = 0., dropout = 0., **kwargs):
         super().__init__()
         assert dim % heads == 0, 'dimensions must be divisible by number of heads'
 
@@ -430,8 +429,9 @@ class LSHSelfAttention(nn.Module):
         self.to_out = nn.Linear(dim, dim)
 
         self.bucket_size = bucket_size
-        self.lsh_attn = LSHAttention(bucket_size=bucket_size, n_hashes=n_hashes, causal=causal, random_rotations_per_head=random_rotations_per_head, attend_across_buckets = attend_across_buckets,  allow_duplicate_attention = allow_duplicate_attention, return_attn = return_attn, **kwargs)
-        self.full_attn = FullQKAttention(causal = causal)
+        self.lsh_attn = LSHAttention(bucket_size=bucket_size, n_hashes=n_hashes, n_local_attn_hashes = n_local_attn_hashes, causal=causal, random_rotations_per_head=random_rotations_per_head, attend_across_buckets = attend_across_buckets,  allow_duplicate_attention = allow_duplicate_attention, return_attn = return_attn, dropout = dropout, **kwargs)
+        self.full_attn = FullQKAttention(causal = causal, dropout = dropout)
+        self.post_attn_dropout = nn.Dropout(post_attn_dropout)
 
         self.use_full_attn = use_full_attn
         self.full_attn_thres = default(full_attn_thres, bucket_size)
@@ -485,13 +485,15 @@ class LSHSelfAttention(nn.Module):
         attn_fn = self.lsh_attn if not use_full_attn else self.full_attn
         partial_attn_fn = partial(attn_fn, query_len = t)
         attn_fn_in_chunks = process_inputs_chunk(partial_attn_fn, chunks = self.attn_chunks)
+
         out, attn, buckets = attn_fn_in_chunks(qk, v, **masks)
         out = split_heads(out).view(b, t, e)
 
         if self.callback is not None:
             self.callback(attn.reshape(b, h, t, -1), buckets.reshape(b, h, -1))
 
-        return self.to_out(out)
+        out = self.to_out(out)
+        return self.post_attn_dropout(out)
 
 # feed forward
 
@@ -573,16 +575,18 @@ class AxialPositionalEncoding(nn.Module):
 # reformer lm
 
 class Reformer(nn.Module):
-    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 8, ff_chunks = 100, attn_chunks = None, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., layer_dropout = 0., lsh_attend_across_buckets = True, lsh_allow_duplicate_attention = True, random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False):
+    def __init__(self, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 8, n_local_attn_hashes = 0, ff_chunks = 100, attn_chunks = None, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., post_attn_dropout = 0., layer_dropout = 0., lsh_attend_across_buckets = True, lsh_allow_duplicate_attention = True, random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False):
         super().__init__()
         self.dim = dim
         self.depth = depth
 
         self.bucket_size = bucket_size
         self.num_mem_kv = num_mem_kv
+
+        self.twin_attention = twin_attention
         self.full_attn_thres = full_attn_thres
 
-        get_attn = lambda: SettableArgs(LSHSelfAttention(dim, heads, bucket_size, n_hashes, causal = causal, dropout = lsh_dropout, attn_chunks = attn_chunks, allow_duplicate_attention = lsh_allow_duplicate_attention, attend_across_buckets = lsh_attend_across_buckets, random_rotations_per_head = random_rotations_per_head, num_mem_kv = num_mem_kv, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, one_value_head = one_value_head))
+        get_attn = lambda: LSHSelfAttention(dim, heads, bucket_size, n_hashes, n_local_attn_hashes = n_local_attn_hashes, causal = causal, dropout = lsh_dropout, post_attn_dropout = post_attn_dropout, attn_chunks = attn_chunks, allow_duplicate_attention = lsh_allow_duplicate_attention, attend_across_buckets = lsh_attend_across_buckets, random_rotations_per_head = random_rotations_per_head, num_mem_kv = num_mem_kv, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, one_value_head = one_value_head)
         get_ff = lambda: FeedForward(dim, dropout = ff_dropout)
 
         if weight_tie:
@@ -605,20 +609,15 @@ class Reformer(nn.Module):
             blocks.append(nn.ModuleList([f, g]))
 
         self.layers = ReversibleSequence(nn.ModuleList(blocks), layer_dropout = layer_dropout, reverse_thres = reverse_thres)
-        self.settables = list(filter(lambda x: isinstance(x, SettableArgs), self.layers.modules()))
-
-    def set_reversible_args(self, *args, **kwargs):
-        for module in self.settables:
-            module.set_args(*args, **kwargs)
 
     def forward(self, x, **kwargs):
         x = torch.cat([x, x], dim = -1)
-        self.set_reversible_args(**kwargs)
-        x = self.layers(x)
+        arg_route = (True, self.twin_attention)
+        x = self.layers(x, arg_route = arg_route, **kwargs)
         return torch.stack(x.chunk(2, dim=-1)).sum(dim=0)
 
 class ReformerLM(nn.Module):
-    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 4, ff_chunks = 100, attn_chunks = 1, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., layer_dropout = 0., random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, emb_dim = None, return_embeddings = False, fixed_position_emb = False, axial_position_emb = False, axial_position_shape = (), axial_position_dims = ()):
+    def __init__(self, num_tokens, dim, depth, max_seq_len, heads = 8, bucket_size = 64, n_hashes = 4, n_local_attn_hashes = 0, ff_chunks = 100, attn_chunks = 1, causal = False, weight_tie = False, lsh_dropout = 0., ff_dropout = 0., post_attn_dropout = 0., layer_dropout = 0., random_rotations_per_head = False, twin_attention = False, use_scale_norm = False, use_full_attn = False, full_attn_thres = 0, reverse_thres = 0, num_mem_kv = 0, one_value_head = False, emb_dim = None, return_embeddings = False, fixed_position_emb = False, axial_position_emb = False, axial_position_shape = (), axial_position_dims = ()):
         super().__init__()
         emb_dim = default(emb_dim, dim)
         self.max_seq_len = max_seq_len
@@ -633,7 +632,7 @@ class ReformerLM(nn.Module):
         else:
             self.pos_emb = AbsolutePositionalEmbedding(emb_dim, max_seq_len)
 
-        self.reformer = Reformer(dim, depth, max_seq_len, heads = heads, bucket_size = bucket_size, n_hashes = n_hashes, ff_chunks = ff_chunks, attn_chunks = attn_chunks, causal = causal, weight_tie = weight_tie, lsh_dropout = lsh_dropout, ff_dropout = ff_dropout, layer_dropout = layer_dropout, random_rotations_per_head = random_rotations_per_head, twin_attention = twin_attention, use_scale_norm = use_scale_norm, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, reverse_thres = reverse_thres, num_mem_kv = num_mem_kv, one_value_head = one_value_head)
+        self.reformer = Reformer(dim, depth, max_seq_len, heads = heads, bucket_size = bucket_size, n_hashes = n_hashes, n_local_attn_hashes = n_local_attn_hashes, ff_chunks = ff_chunks, attn_chunks = attn_chunks, causal = causal, weight_tie = weight_tie, lsh_dropout = lsh_dropout, ff_dropout = ff_dropout, post_attn_dropout = 0., layer_dropout = layer_dropout, random_rotations_per_head = random_rotations_per_head, twin_attention = twin_attention, use_scale_norm = use_scale_norm, use_full_attn = use_full_attn, full_attn_thres = full_attn_thres, reverse_thres = reverse_thres, num_mem_kv = num_mem_kv, one_value_head = one_value_head)
         self.to_logits = identity if return_embeddings else nn.Linear(dim, num_tokens)
 
     def forward(self, x, **kwargs):
